#!/usr/bin/env ruby
#
# pem2csv.rb v0.9($Revision: 1.1.1.1 $) $Date: 2002/04/10 14:48:37 $
#
# You may redistribute it and/or modify it under the same license
# terms as Ruby.
#
# Documents and latest version of `pem2csv.rb' are found at:
#    <URL:http://www.ish.ic.kanagawa-it.ac.jp/~katsuwo/software/pem2csv/>
#
# Copyright (C) 2002 katsuwo(Takeshi Komiya)
#

require 'singleton'

module PalmDatabase
  class Time < ::Time
    @@differs = self.gm(1970 + (1970 - 1904)).to_i

    def self.at(sec)
      sec -= @@differs
      super(sec)
    end
  end
  class DatabaseHeader
    attr_reader(:name, :flags, :version, :ctime, :mtime, :btime, :modified,
                :appinfosize, :sortinfosize, :type, :creatorid, :idseed,
                :recordlist, :records)

    def initialize(str)
      array = str.unpack('A32nnNNNNNNa4a4NNn')

      @name         = array[0]
      @flags        = array[1]
      @version      = array[2]
      @ctime        = Time.at(array[3])
      @mtime        = Time.at(array[4])
      @btime        = Time.at(array[5])
      @modified     = array[6]
      @appinfosize  = array[7] 
      @sortinfosize = array[8]
      @type         = array[9]
      @cretorid     = array[10]
      @idseed       = array[11]
      @recordlist   = array[12]
      @records      = array[13]
    end
  end
  class RecordHeader
    attr_reader(:offset, :attr, :uniqueid)

    def initialize(str)
      array = str.unpack('Ncccc')

      @offset   = array[0]
      @attr     = array[1].to_i
#      @uniqueid = array[2..4]
    end
    def private?
      (@attr & 0x10).nonzero?
    end
    def category
      @attr & 0x0f
    end
  end

  class DatabaseParser
    class Configuration
      include Singleton
      attr_accessor(:readPrivate)

      def initialize
        @readPrivate = false
      end
    end
    class Record
      attr_accessor(:header)

      def initialize(header)
        @header = header
      end
      def category
        @header.category
      end
      def private?
        @header.private?
      end
    end

    attr_reader(:records)

    # load configuration.
    def self.option
      Configuration.instance
    end
    def option
      Configuration.instance
    end

    def initialize(stream)
      @recordheaders = []
      @records = []

      # read DatabaseHeader.
      str = stream.read(78)
      hdr = PalmDatabase::DatabaseHeader.new(str)

      # read RecordHeaders.
      hdr.records.times do
        str = stream.read(8)
        elem = PalmDatabase::RecordHeader.new(str)

        @recordheaders.push(elem)
      end

      # avoid pad:p
      stream.read(2)

      # read AppInfo.
      str = stream.read(hdr.appinfosize)
      parseAppInfo(str)

      # read all records.
      @recordheaders.each_with_index do |header, index|
        next if(header.private? && (not option.readPrivate))

        stream.seek(header.offset)

        if(@recordheaders[index + 1])
          str = stream.read(@recordheaders[index + 1].offset - header.offset)
        else
          str = stream.read
        end

        # create payment object.
        record = generateRecord(header)
        parseRecord(str, record)
        @records.push(record)
      end
    end
  end
  class PEMDatabase < DatabaseParser
    class PEMRecord < DatabaseParser::Record
      attr_accessor(:header, :date, :amount, :purpose)
    end

    attr_reader(:categories)
    def parseAppInfo(str)
      @categories = str[2..-1].unpack('A16' * 16)
      @categories.each{|x| x.gsub!(/\000/, '')}
    end
    def parseRecord(str, record)
      ary = str.unpack('nNx2A*')
      year  = 1904 + (ary[0] >> 9)
      month = (ary[0] & 0x01e0) >> 5
      day   = (ary[0] & 0x001f)

      record.date    = Time.gm(year, month, day)
      record.amount  = [ary[1]].pack('l').unpack('l').shift
      record.purpose = ary[2]
    end
    def generateRecord(header)
      PEMRecord.new(header)
    end
  end
end

def usage
  print <<USAGE
  usage: #{File::basename($0)} [Options] [FILE]

  Options:
    -h, --help        show this usage.
    -a, --all         print all records in database.
    -p, --private     print private records in database.
    -e ENCODING, --encoding ENCODING
                      set output encoding to ENCODING.
                      ENCODING can be taken 'sjis' or 'euc'.
                      default encoding is Shift-JIS.

USAGE
end

module PEM2CSV
  class Option
    include Singleton
    attr_accessor :showprivate, :encoding

    def self.parse
      parser = GetoptLong.new(
                   ['--all',      '-a', GetoptLong::NO_ARGUMENT],
                   ['--encoding', '-e', GetoptLong::REQUIRED_ARGUMENT],
                   ['--help',     '-h', GetoptLong::NO_ARGUMENT],
                   ['--private',  '-p', GetoptLong::NO_ARGUMENT])

      parser.each_option do |name, arg|
        case name
          when '--all'
            PalmDatabase::DatabaseParser::option.readPrivate = true
          when '--encoding'
            if(arg =~ /(sjis|euc)/i)
              instance.encoding = arg
            else
              raise GetoptLong::InvalidOption.new('Unknown encoding: ' + arg)
            end
          when '--help'
            usage
            exit
          when '--private'
            PalmDatabase::DatabaseParser::option.readPrivate = true
            instance.showprivate = true
        end
      end

      instance
    end
    def initialize
      @showprivate = false
    end
  end
end

if($0 == __FILE__)
  require 'getoptlong'
  require 'nkf'

  PEM2CSV::Option.parse

  if(ARGV.size != 1)
    usage
    exit
  end

  # parse database.
  filename = ARGV.shift
  pemdb = File::open(filename, 'rb') do |fd|
    PalmDatabase::PEMDatabase.new(fd)
  end

  # formatting.
  pemdb.records.sort{|x,y| x.date <=> y.date}.each do |payment|
    next if(PEM2CSV::Option.instance.showprivate && (not payment.private?))

    # convert encodings.
    case PEM2CSV::Option.instance.encoding
      when 'sjis'
        purpose = NKF.nkf('-s', payment.purpose)
        catname = NKF.nkf('-s', pemdb.categories[payment.category])
      when 'euc'
        purpose = NKF.nkf('-e', payment.purpose)
        catname = NKF.nkf('-e', pemdb.categories[payment.category])
      else
        purpose = payment.purpose
        catname = pemdb.categories[payment.category]
    end

    printf "%s,\"%s\",\"%s\",%d\n", payment.date.strftime('%Y/%m/%d'), 
                                    catname, purpose, payment.amount
  end
end
